@@ -1,10 +1,12 @@
 import re
 from typing import Iterable
 from psycopg import sql, connect
+from util import iterable_difference
 
 
+# TODO: today use date() to type cast; change regex to '::' for type cast vs '()' for true funcs
 # Regex to seperate sql_func from column name
-__compiled_pattern = re.compile(pattern=r"\((\w*)\)")
+__compiled_pattern = re.compile(pattern=r"(\w*)\((\w*)\)")
 
 
 # =================== Snippet Util Functions ===================
@@ -23,56 +25,43 @@ def check_for_func(sequence: Iterable) -> bool:
         bool: True if function found
     """
     # make all elements strings
-    it = map(str, sequence)
-    combined = "".join(it)
-    return "(" in combined
+    seq_of_str = map(str, sequence)
+    # combine seq of str into one long str
+    combined_seq = "".join(seq_of_str)
+
+    # return true or false
+    return "(" in combined_seq
 
 
-def extract_sqlfunc_colname(column_name: str):
-    """If sql function passed with column name,
-    seperate function from column name.
-    ie Date(ts) -> (Date(, ts)
-
-    Args:
-        column_name (str): _description_
-
-    Returns:
-        str or tuple: column_name or tuple of func, column_name
-    """
-
-    # pattern = r"\((\w*)\)"
-    result = re.search(__compiled_pattern, column_name)
-
-    # Extract matching values of all groups
-    if result:
-        column = result.group(1)
-        func = column_name.replace(column + ")", "").upper()
-        return func, column
-
-    # if no '(' return original column name
-    return column_name
-
-
-def colname_snip(column_detail: str | tuple):
+def colname_snip(sqlfunc_colname: tuple):
     """Return escaped sql snippet, accomodate column names
     wrapped by sql functions.
 
     Args:
-        column_detail (str | tuple): column_name as str or tuple (sql_func, column_name)
+        column_detail (str | tuple): column_name as str
+        or tuple (sql_func, column_name)
 
     Returns:
         Composed: snippet of sql statment
     """
-    # return escaped column name
-    if isinstance(column_detail, str):
+    sqlfunc, colname = sqlfunc_colname
+
+    # return snippet of sql func wrapping column name
+    if sqlfunc is None:
+        # return escaped column name
         return sql.SQL("{}").format(
-            sql.Identifier(column_detail),
+            sql.Identifier(colname),
         )
 
-    # else if tuple return snippet if sql func wrapping column
-    return sql.SQL("{}{})").format(
-        sql.SQL(column_detail[0]),
-        sql.Identifier(column_detail[1]),
+    # Returns sql func format; Keep for later use for true functions
+    # return sql.SQL("{}({})").format(
+    #     sql.SQL(sqlfunc),
+    #     sql.Identifier(colname),
+    # )
+
+    return sql.SQL("{}::{}").format(
+        sql.Identifier(colname),
+        sql.SQL(sqlfunc),
     )
 
 
@@ -82,23 +71,23 @@ def colname_snip(column_detail: str | tuple):
 def create_unique_index(table, keys):
 
     # Note name will include parenthsis if passed
-    # unique index name
+    # in unique index name
     uix_name = f'{table}_{"_".join(keys)}_uix'
 
-    # if sql function in the any key
+    # If sql function in the any key
     if check_for_func(keys):
 
-        # Return tuple of functions and key else just key
-        seperated_keys = map(extract_sqlfunc_colname, keys)
+        # Return tuple of (sql func, colname)
+        sqlfunc_keys = map(get_sqlfunc_colname, keys)
 
         # sql snippet to create unique index
         return sql.SQL("CREATE UNIQUE INDEX {} ON {} ({});").format(
             sql.Identifier(uix_name),
             sql.Identifier(table),
-            sql.SQL(", ").join(map(colname_snip, seperated_keys)),
+            sql.SQL(", ").join(map(colname_snip, sqlfunc_keys)),
         )
 
-    # sql snippet to create unique index
+    # Sql snippet to create unique index
     return sql.SQL("CREATE UNIQUE INDEX {} ON {} ({});").format(
         sql.Identifier(uix_name),
         sql.Identifier(table),
@@ -116,28 +105,38 @@ def exclude_sql(col):
     )
 
 
-def upsert_snip(table: str, columns, keys):
+def upsert_snip(
+    table: str, columns: Iterable, keys: Iterable, exclude_update: Iterable = None
+):
+
+    update_columns = columns
+
+    # if exclude columns from update then determine update_columns
+    if exclude_update:
+        update_columns = iterable_difference(columns, exclude_update)
 
     # if sql function in the any key
     if check_for_func(keys):
 
-        # Return tuple of (function, key) else just key
-        func_keys = map(extract_sqlfunc_colname, keys)
+        # Return tuple of (sql func, key)
+        sqlfunc_keys = map(get_sqlfunc_colname, keys)
 
+        # Sql snippet to upsert
         return sql.SQL(
-            "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {}"
+            "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {};"
         ).format(
             sql.Identifier(table),
             sql.SQL(", ").join(map(sql.Identifier, columns)),
             sql.SQL(", ").join(map(sql.Placeholder, columns)),
             # conflict target
-            sql.SQL(", ").join(map(colname_snip, func_keys)),
+            sql.SQL(", ").join(map(colname_snip, sqlfunc_keys)),
             # set new values
-            sql.SQL(", ").join(map(exclude_sql, columns)),
+            sql.SQL(", ").join(map(exclude_sql, update_columns)),
         )
 
+    # Sql snippet to upsert
     return sql.SQL(
-        "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {}"
+        "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {};"
     ).format(
         sql.Identifier(table),
         sql.SQL(", ").join(map(sql.Identifier, columns)),
@@ -145,7 +144,7 @@ def upsert_snip(table: str, columns, keys):
         # conflict target
         sql.SQL(", ").join(map(sql.Identifier, keys)),
         # set new values
-        sql.SQL(", ").join(map(exclude_sql, columns)),
+        sql.SQL(", ").join(map(exclude_sql, update_columns)),
     )
 
 
@@ -154,12 +153,13 @@ def upsert_snip(table: str, columns, keys):
 
 def insert_ignore_snip(table: str, columns, keys):
 
-    # if sql function in the any key
+    # If sql function in the any key
     if check_for_func(keys):
 
-        # Return tuple of functions and key else just key
-        seperated_keys = map(extract_sqlfunc_colname, keys)
+        # Return tuple (sql func, key)
+        sqlfunc_keys = map(get_sqlfunc_colname, keys)
 
+        # Sql snippet to insert ignore
         return sql.SQL(
             "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO NOTHING"
         ).format(
@@ -167,9 +167,10 @@ def insert_ignore_snip(table: str, columns, keys):
             sql.SQL(", ").join(map(sql.Identifier, columns)),
             sql.SQL(", ").join(map(sql.Placeholder, columns)),
             # conflict target
-            sql.SQL(", ").join(map(colname_snip, seperated_keys)),
+            sql.SQL(", ").join(map(colname_snip, sqlfunc_keys)),
         )
 
+    # Sql snippet to insert ignore
     return sql.SQL(
         "INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO NOTHING"
     ).format(
@@ -185,7 +186,7 @@ def insert_ignore_snip(table: str, columns, keys):
 
 # Note: Seperated from where_snip() to possibly reuse
 # on other dictionaries needing to be composed for other
-# snippets in future 
+# snippets in future
 def compose_key_value(key_value: tuple) -> tuple:
     """Take key_value tuple from dictionary via .items()
     and create a composed key_value tuple for use in creating
@@ -206,15 +207,17 @@ def compose_key_value(key_value: tuple) -> tuple:
 
     # Check if colname has a function
     if "(" in colname:
-        # pattern = r"\((\w*)\)"
+        # pattern = r"(\w*)\((\w*)\)"
         result = re.search(__compiled_pattern, colname)
 
         # Extract matching values of all groups
         # this is done to escape column name
-        column = result.group(1)
-        func = colname.replace(column + ")", "").upper()
+        sqlfunc = result.group(1).upper()
+        column = result.group(2)
 
-        composed_column = sql.SQL("{}{})").format(sql.SQL(func), sql.Identifier(column))
+        composed_column = sql.SQL("{}({})").format(
+            sql.SQL(sqlfunc), sql.Identifier(column)
+        )
 
         return composed_column, composed_value
 
@@ -223,10 +226,8 @@ def compose_key_value(key_value: tuple) -> tuple:
     return composed_column, composed_value
 
 
-
-
 # Note: Seperated from compose_key_value() as there
-# may be need for compose other dictionaries in future 
+# may be need for compose other dictionaries in future
 def where_snip(colname_value: tuple):
     """Represent where clause colname=value
 
@@ -248,7 +249,7 @@ def delete_snip(table: str, where: dict):
 
     Args:
         table (str): database table name
-        where (dict): dict(colname=value) that 
+        where (dict): dict(colname=value) that
         filters value to remove from table
 
     Returns:
@@ -264,10 +265,112 @@ def delete_snip(table: str, where: dict):
     )
 
 
+# =================== Update Snippets ===================
+
+
+def get_sqlfunc_colname(colname: str):
+    """
+    Extract sql function from column name.
+    Return (sql function, column name)
+    If no sql function return None in tuple.
+    Used on Iterable of strings via map().
+
+    Args:
+        colname (str): column name
+
+    Returns:
+        tuple: (None, colname) or (sqlfunc, colname)
+    """
+    # Check if colname has a function
+    if "(" in colname:
+        try:
+            # pattern = r"(\w*)\((\w*)\)"
+            result = re.search(__compiled_pattern, colname)
+
+            # Extract matching values of all groups
+            # this is done to escape column name
+            sqlfunc = result.group(1).upper()
+            colname = result.group(2)
+
+            return sqlfunc, colname
+        except:
+            print("Invalid column name passed!")
+    return None, colname
+
+
+# function used to map passed dictionary values to column names
+# TODO: Truly distinguish between sql func and type cast; limited use today
+def colname_placeholder_snip(sqlfunc_colname: tuple):
+
+    """
+    Return sql snip for where and set clauses with
+    named placeholder for value;
+    ex. "SET col=value, col2=value2
+    colname=%(colname)s or sqlfunc(colname)=%(colname)s
+
+    Args:
+        sqlfunc_colname (tuple): (sqlfunc, colname)
+
+    Returns:
+        Composable: colname=%(colname)s
+    """
+    # unpack sqlfunc_colname to determine colname placeholder syntax
+    sqlfunc, colname = sqlfunc_colname
+    
+    if sqlfunc is None:
+        return sql.SQL("{}={}").format(
+            sql.Identifier(colname),
+            sql.Placeholder(colname),
+        )
+
+    # NOTE: KEEP FOR NOW IN CASE WANT TO PROCESS TRUE FUNCTIONS VS TYPE CAST FUNCTIONS
+    # return sql.SQL("{}({})={}").format(
+    #     sql.SQL(sqlfunc),
+    #     sql.Identifier(colname),
+    #     sql.Placeholder(colname),
+    # )
+
+    # type cast if func found, Update future to handle better
+    return sql.SQL("{}::{}={}::{}").format(
+        sql.Identifier(colname),
+        sql.SQL(sqlfunc),
+        sql.Placeholder(colname),
+        sql.SQL(sqlfunc),
+    )
+
+
+def update_snip(
+    table: str, columns: Iterable, keys: Iterable, exclude_update: Iterable = None
+):
+
+    # if exclude columns from update then determine update_columns
+    if exclude_update:
+        columns = iterable_difference(columns, exclude_update)
+
+    columns = map(get_sqlfunc_colname, columns)
+    keys = map(get_sqlfunc_colname, keys)
+
+    return sql.SQL("UPDATE {} SET {} WHERE {};").format(
+        sql.Identifier(table),
+        sql.SQL(", ").join(map(colname_placeholder_snip, columns)),
+        sql.SQL(" AND ").join(map(colname_placeholder_snip, keys)),
+    )
+
+
+def insert_snip(table: str, columns: Iterable):
+
+    return sql.SQL("INSERT INTO {} ({}) VALUES ({})").format(
+        sql.Identifier(table),
+        sql.SQL(", ").join(map(sql.Identifier, columns)),
+        sql.SQL(", ").join(map(sql.Placeholder, columns)),
+    )
+
+
 if __name__ == "__main__":
     # dev testing, remove later
     import os
     from timeit import timeit
+
     # print("SNIPPET.PY")
 
     conn_import: dict = {
@@ -284,24 +387,24 @@ if __name__ == "__main__":
         dtest = {"name": "bos", "age": 34, "Date(ts)": "2022-04-19"}
         # dtest = {"name": "bos", "age": 34, "ts": "2022-04-19"}
 
-        
-
         # tests
-        # snipp = extract_sqlfunc_colname("Date(ts)")
         # snipp = create_unique_index("mytable", ["name", "Date(ts)"])
-        
+
         # timeit(delete_snip, "mytable", dtest)
         # def a():
-            # snipp = delete_snip("mytable", dtest)
-            # print(snipp.as_string(conn))
+        # snipp = delete_snip("mytable", dtest)
+        # print(snipp.as_string(conn))
 
-        snipp = delete_snip("mytable", dtest)
+        # snipp = delete_snip("mytable", dtest)
+
+        snipp = update_snip(
+            table="mytable", columns=["a", "b", "c"], keys=["Date(b)", "d"]
+        )
         print(snipp.as_string(conn))
 
         # Code used to time speed of functions
         # print(timeit(stmt='delete_snip("mytable", dtest)', number=1000, globals=globals())/1000)
         # print(timeit(stmt='delete_snip("mytable", dtest)', setup='from __main__ import delete_snip, dtest', number=1))
         # print(timeit(stmt='a()', setup='from __main__ import a', number=100))
-        
 
         conn.close()
