@@ -15,12 +15,10 @@ from wrapg import util, snippet
 # Inspired by dataset library that wraps sqlalchemy
 # ================================= TODO ================================
 # TODO: Can query() return explain analyse info?
-# TODO: Conditionally import config?
-# TODO: Implement delete
 # ===========================================================================
 
 # TODO: Add proper exceptions if parameters are missing or do not work
-# TODO: expose conn_import in init.py as class to easily modify attributes?? vs dict?
+# TODO: Expose conn_import in init.py as class to easily modify attributes?? vs dict?
 conn_import: dict = {
     "user": os.environ.get("PG_USER"),
     "password": os.environ.get("PG_PASSWORD"),
@@ -30,7 +28,7 @@ conn_import: dict = {
 }
 
 
-def query(raw_sql: str, to_df=False, conn_kwargs: dict = None):
+def query(raw_sql: str, to_df: bool = False, conn_kwargs: dict = None):
     """Function to send raw sql query to postgres db.
 
     Args:
@@ -112,31 +110,27 @@ def insert(data: Iterable[dict] | pd.DataFrame, table: str, conn_kwargs: dict =
 
             if uniform == 1:
                 # Dynamic insert query for dictionaries
-                qry = sql.SQL("INSERT INTO {} ({}) VALUES ({})").format(
-                    sql.Identifier(table),
-                    sql.SQL(", ").join(map(sql.Identifier, columns)),
-                    sql.SQL(", ").join(map(sql.Placeholder, columns)),
-                    # Dynamic insert query for list of tuples
-                    # sql.SQL(", ").join(sql.Placeholder() * len(columns)),
-                )
-                # print(qry.as_string(conn))
+                insert_qry = snippet.insert_snip(table=table, columns=columns)
+                # print(insert_qry.as_string(conn))
+
                 # One insert for all dictionaries
-                cur.executemany(query=qry, params_seq=rows)
+                cur.executemany(query=insert_qry, params_seq=rows)
+
+                # Return # of inserted records
+                return cur.rowcount
 
             else:
+                rw_count = 0
                 # For non uniform data
                 for row in rows:
                     # Dynamic insert query for dictionaries
-                    qry = sql.SQL("INSERT INTO {} ({}) VALUES ({})").format(
-                        sql.Identifier(table),
-                        sql.SQL(", ").join(map(sql.Identifier, tuple(row))),
-                        sql.SQL(", ").join(map(sql.Placeholder, tuple(row))),
-                    )
-                    # Seperate insert for each dictionary
-                    cur.execute(query=qry, params=row)
+                    insert_qry = snippet.insert_snip(table=table, columns=tuple(row))
 
-            # Make the changes to the database persistent
-            conn.commit()
+                    # Seperate insert for each dictionary
+                    cur.execute(query=insert_qry, params=row)
+                    rw_count += cur.rowcount
+
+                return rw_count
 
 
 def insert_ignore(
@@ -194,6 +188,9 @@ def insert_ignore(
                     # print(qry.as_string(conn))
                     cur.executemany(query=qry, params_seq=rows)
 
+                    # TODO: add return count and test for all cases
+                    # return cur.rowcount
+
                 else:
                     for row in rows:
                         # get sql qry based on passed parameters for each row
@@ -246,7 +243,9 @@ def insert_ignore(
 def upsert(
     data: Iterable[dict] | pd.DataFrame,
     table: str,
-    keys: list,
+    keys: Iterable,
+    exclude_update: Iterable = None,
+    use_index: bool = True,
     conn_kwargs: dict = None,
 ):
     """Function for SQL's INSERT ON CONFLICT DO UPDATE SET
@@ -258,7 +257,9 @@ def upsert(
     Args:
         data (Iterable[dict] | pd.DataFrame): data in form of dict, list of dict, or dataframe
         table (str): name of database table
-        keys (list): Iterable of columns
+        keys (Iterable): column names used to filter & match records; synonymous with sql WHERE
+        use_index (bool): if False first try to update then insert without use of an index.
+        exclude_update (Iterable): exclude columns from updating database
         conn_kwargs (dict, optional): Specify/overide conn kwargs. See full list of options,
         https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS.
         Defaults to None, recommend importing via .env file.
@@ -289,87 +290,175 @@ def upsert(
             # DO UPDATE SET email=excluded.email
             # WHERE ...;
 
-            try:
-                if uniform == 1:
-                    qry = snippet.upsert_snip(table=table, columns=columns, keys=keys)
-                    # print(qry.as_string(conn))
-                    cur.executemany(query=qry, params_seq=rows)
-
-                else:
-                    for row in rows:
-                        # Note tupe(row) returns column keys for each record
-                        qry = snippet.upsert_snip(
-                            table=table, columns=tuple(row), keys=keys
-                        )
-                        # print(qry.as_string(conn))
-                        cur.execute(query=qry, params=row)
-
-            # Catch no unique index error
-            except errors.InvalidColumnReference as e:
-                print(">>> Error: ", e)
-                print(f"> Creating unique index for {keys}...")
-                # !important, cannot attempt other operations after error unless rollback()
-                conn.rollback()
-
+            if use_index is True:
                 try:
-                    # Create unique index & try upsert again
-                    uix_sql = snippet.create_unique_index(table=table, keys=keys)
-                    print(uix_sql.as_string(conn))
-                    cur.execute(query=uix_sql)
-
+                    # Process uniform data
                     if uniform == 1:
                         qry = snippet.upsert_snip(
-                            table=table, columns=columns, keys=keys
+                            table=table,
+                            columns=columns,
+                            keys=keys,
+                            exclude_update=exclude_update,
                         )
                         # print(qry.as_string(conn))
                         cur.executemany(query=qry, params_seq=rows)
+                        return cur.rowcount
 
+                    # Process Non-Uniform Data
                     else:
+                        rw_count = 0
                         for row in rows:
                             # Note tupe(row) returns column keys for each record
                             qry = snippet.upsert_snip(
-                                table=table, columns=tuple(row), keys=keys
+                                table=table,
+                                columns=tuple(row),
+                                keys=keys,
+                                exclude_update=exclude_update,
                             )
                             # print(qry.as_string(conn))
                             cur.execute(query=qry, params=row)
+                            rw_count += cur.rowcount
+                        return rw_count
 
-                except Exception as indx_error:
-                    print(">>> Error: ", indx_error)
+                # Catch no unique index error
+                # TODO: Can i check if index exist rather than using error
+                except errors.InvalidColumnReference as e:
+                    # print(">>> Error: ", e)
+                    # print(f"> Creating unique index for {keys}...")
+                    # !Important, cannot attempt other operations after error unless rollback()
+                    conn.rollback()
+
+                    # Create unique index & try upsert again
+                    try:
+                        uix_sql = snippet.create_unique_index(table=table, keys=keys)
+                        # print(uix_sql.as_string(conn))
+                        cur.execute(query=uix_sql)
+
+                        # Process Uniform data
+                        if uniform == 1:
+                            qry = snippet.upsert_snip(
+                                table=table,
+                                columns=columns,
+                                keys=keys,
+                                exclude_update=exclude_update,
+                            )
+                            # print(qry.as_string(conn))
+                            cur.executemany(query=qry, params_seq=rows)
+                            return cur.rowcount
+
+                        # Process Non-uniform data
+                        else:
+                            rw_count = 0
+                            for row in rows:
+                                # Note tupe(row) returns column keys for each record
+                                qry = snippet.upsert_snip(
+                                    table=table,
+                                    columns=tuple(row),
+                                    keys=keys,
+                                    exclude_update=exclude_update,
+                                )
+                                # print(qry.as_string(conn))
+                                cur.execute(query=qry, params=row)
+                                rw_count += cur.rowcount
+
+                            return rw_count
+
+                    except Exception as indx_error:
+                        print(">>> Error: ", indx_error)
+                        quit()
+
+                # Handle all other exceptions
+                except Exception as ee:
+                    print("Exception: ", ee.__init__)
                     quit()
 
-            # Handle all other exceptions
-            except Exception as ee:
-                print("Exception: ", ee.__init__)
-                quit()
+            if use_index is False:
 
-            # Make the changes to the database persistent
-            conn.commit()
+                # If use_index is False
+                # First attempt to update data, confirm result
+                # If no update, then insert record
 
+                # Process uniform data without index
+                if uniform == 1:
 
-"""
-# TODO: **upsert without creating auto-index or constriants
-**Add as option to exisiting upsert(auto_uix=False)
-def upsert_wo_idx():
-    Sudo code
-    This function would be run inside a for loop within a transaction
-    every iteration would result in a update or an insert
+                    # Update qry for uniform data
+                    update_qry = snippet.update_snip(
+                        table=table,
+                        columns=columns,
+                        keys=keys,
+                        exclude_update=exclude_update,
+                    )
+                    cur.executemany(query=update_qry, params_seq=rows)
+
+                    # if all records updated, finish return # of updated records
+                    if cur.rowcount == len(rows):
+                        # print("All records updated, uniform")
+                        return cur.rowcount
+
+                    # Dynamic insert query for dictionaries
+                    insert_qry = snippet.insert_snip(table=table, columns=columns)
+
+                    # Insert all records (no records can be updated)
+                    if cur.rowcount == 0:
+                        # One insert for all dictionaries
+                        cur.executemany(query=insert_qry, params_seq=rows)
+                        # print("All records inserted, uniform")
+
+                        return cur.rowcount
+
+                    # Update or Insert new records need to be inserted
+                    rw_count = 0
+                    # print("Records updated & inserted, uniform")
+                    for row in rows:
+                        # Update single record
+                        cur.execute(query=update_qry, params=row)
+                        rw_count += cur.rowcount
+
+                        # if no update then insert record
+                        if cur.rowcount == 0:
+                            cur.execute(query=insert_qry, params=row)
+                            rw_count += cur.rowcount
+
+                    # total records updated or inserted
+                    return rw_count
+
+                # Process Non-uniform data without index
+                else:
+                    # print("Records updated & inserted, Non-uniform")
+                    rw_count = 0
+                    for row in rows:
+                        # Update qry for uniform data
+                        update_qry = snippet.update_snip(
+                            table=table,
+                            columns=tuple(row),
+                            keys=keys,
+                            exclude_update=exclude_update,
+                        )
+                        # print(update_qry.as_string(conn))
+                        cur.execute(query=update_qry, params=row)
+                        rw_count += cur.rowcount
+
+                        # if no update then insert record
+                        if cur.rowcount == 0:
+
+                            # Dynamic insert query for dictionaries
+                            insert_qry = snippet.insert_snip(
+                                table=table, columns=tuple(row)
+                            )
+                            cur.execute(query=insert_qry, params=row)
+
+                            rw_count += cur.rowcount
+                    return rw_count
 
-    try:
-        update() on keys
-        if update() return value is 0
-        then insert()
-    this code should be much slower but not require creating
-    unique constriants which may be hard to manage especially
-    for novice users
-"""
 
 # ================================= UPDATE Function ================================
 def update(
     data: list[dict] | pd.DataFrame,
     table: str,
     keys: Iterable,
+    exclude_update: Iterable = None,
     conn_kwargs: dict = None,
-):
+) -> int:
     """Function for SQL's UPDATE
 
     If rows with matching keys exist, update row values.
@@ -379,10 +468,14 @@ def update(
     Args:
         data (list[dict] | pd.DataFrame): data in form of dict, list of dict, or dataframe
         table (str): name of database table
-        keys (Iterable): Iterable of column names synonymous with sql WHERE
+        keys (Iterable): column names used to filter & match records; synonymous with sql WHERE
+        exclude_update (Iterable): exclude columns from updating database
         conn_kwargs (dict, optional): Specify/overide conn kwargs. See full list of options,
         https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS.
         Defaults to None, recommend importing via .env file.
+
+    Returns:
+        int: # of updated records
     """
 
     # Inspect data and return columns and rows
@@ -408,51 +501,40 @@ def update(
             # SET column1 = value1,
             #     column2 = value2,
             #     ...
-            # TODO: Where clause is optional, add code if None update all
             # WHERE column = value, ...
             # RETURNING * | output_expression AS output_name;
 
-            # Function to compose col=value sql str
-            def column_value_str(column_names: Iterable):
-                """Create psycopg composable sql string for
-                variable number of columns/value pairs
-                ie column=value scenerios, col=%(col)s
-
-                Args:
-                    column_names (Iterable): column names
-                """
-                # function used to map to column names
-                def set_sql(col):
-                    return sql.SQL("{}={}").format(
-                        sql.Identifier(col),
-                        sql.Placeholder(col),
-                    )
-
-                return map(set_sql, column_names)
-
             if uniform == 1:
                 # print("> Uniform Data..")
-                qry = sql.SQL("UPDATE {} SET {} WHERE {}").format(
-                    sql.Identifier(table),
-                    sql.SQL(", ").join(column_value_str(columns)),
-                    sql.SQL(", ").join(column_value_str(keys)),
+                qry = snippet.update_snip(
+                    table=table,
+                    columns=columns,
+                    keys=keys,
+                    exclude_update=exclude_update,
                 )
-                # print(qry.as_string(conn))
-
+                print(qry.as_string(conn))
                 cur.executemany(query=qry, params_seq=rows)
+
+                # Return # of updated records
+                return cur.rowcount
+
             else:
                 # print(">> Non-Uniform Data...")
+                rwcount = 0
                 for row in rows:
-                    qry = sql.SQL("UPDATE {} SET {} WHERE {}").format(
-                        sql.Identifier(table),
-                        sql.SQL(", ").join(column_value_str(tuple(row))),
-                        sql.SQL(", ").join(column_value_str(keys)),
+                    qry = snippet.update_snip(
+                        table=table,
+                        columns=tuple(row),
+                        keys=keys,
+                        exclude_update=exclude_update,
                     )
 
+                    print(qry.as_string(conn))
                     cur.execute(query=qry, params=row)
+                    rwcount += cur.rowcount
 
-            # Make the changes to the database persistent
-            conn.commit()
+                # Return # of updated records
+                return rwcount
 
 
 def create_table(table: str, columns: dict, conn_kwargs: dict = None):
@@ -648,7 +730,7 @@ def delete(table: str, where: dict, conn_kwargs: dict = None):
 # ================================= Clear_table Function ================================
 def clear_table(table: str, conn_kwargs: dict = None):
     """!!Caution!! Function for SQL's Delete used to delete 'ALL' records in specified table.
-    
+
     Args:
         table (str): name of database table
         conn_kwargs (dict, optional): Specify/overide conn kwargs. See full list of options,
